/*
The software in this package is distributed under the GNU General
Public License version 2 (with a special exception described below).

A copy of GNU General Public License (GPL) is included in this distribution,
in the file COPYING.GPL.

As a special exception, if other files instantiate templates or use macros
or inline functions from this file, or you compile this file and link it
with other works to produce a work based on this file, this file
does not by itself cause the resulting work to be covered
by the GNU General Public License.

However the source code for this file must still be made available
in accordance with section (3) of the GNU General Public License.

This exception does not invalidate any other reasons why a work based
on this file might be covered by the GNU General Public License.
*/
#include "host/pingscheduler.h"

#include <iostream>
#include <limits>

#include <boost/bind.hpp>

#include <logfunc.hpp>

#include "boost_assert_handler.h"
#include "dns/dnsresolver.h"
#include "host/pingerfactory.h"
#include "icmp/icmppinger.h"
#include "link/linkstatus.h"

using namespace std;
using boost::asio::io_service;
using boost::bind;
using boost::date_time::time_resolution_traits_adapted64_impl;
using boost::posix_time::microsec_clock;
using boost::posix_time::ptime;
using boost::posix_time::seconds;
using boost::shared_ptr;
using I2n::Logger::GlobalLogger;

//-----------------------------------------------------------------------------
// PingScheduler
//-----------------------------------------------------------------------------

/**
 * @brief Parameterized constructor.
 *
 * @param io_serv The one @c io_serv object that controls async processing
 * @param network_interface The name of the network interface originating the pings.
 * @param destination_address The remote address to ping.
 * @param destination_port The remote port to ping.
 * @param ping_protocol_list A list of protocols to use.
 * @param ping_interval_in_sec Amount of time between each ping.
 * @param ping_fail_percentage_limit Maximum amount of pings that can fail.
 * @param nameserver Server to resolve the addresses.
 * @param link_analyzer The object to monitor the link status.
 * @param first_delay Delay in seconds from start_pinging to first ping attempt
 */
PingScheduler::PingScheduler(
        const IoServiceItem io_serv,
        const string &network_interface,
        const string &destination_address,
        const uint16_t destination_port,
        const PingProtocolList &ping_protocol_list,
        const long ping_interval_in_sec,
        const int ping_fail_percentage_limit,
        const string &nameserver,
        const int max_address_resolution_attempts,
        const int ping_reply_timeout,
        const int resolved_ip_ttl_threshold,
        LinkStatusItem link_analyzer,
        const int first_delay

) :
    NextPingTimer( *io_serv ),
    NextAddressTimer( *io_serv ),
    TimeSentLastPing( microsec_clock::universal_time() ),
    PingIntervalInSec( ping_interval_in_sec ),
    AddressResolveIntervalInSec( ping_interval_in_sec ),
    HostAnalyzer( destination_address, ping_fail_percentage_limit, link_analyzer ),
    FirstDelay( first_delay ),
    AddressResolutionAttempts( 0 ),
    MaxAddressResolutionAttempts( max_address_resolution_attempts ),
    EverHadAnyIP( false ),
    Ping()
{
    BOOST_ASSERT( !network_interface.empty() );
    BOOST_ASSERT( !destination_address.empty() );
    BOOST_ASSERT( 0 < destination_port );
    BOOST_ASSERT( 0 < ping_interval_in_sec );
    BOOST_ASSERT( (0 <= ping_fail_percentage_limit) && (ping_fail_percentage_limit <= 100) );
    BOOST_ASSERT( !nameserver.empty() );

    Ping = PingerFactory::createPinger(
            ping_protocol_list,
            io_serv,
            network_interface,
            destination_address,
            destination_port,
            nameserver,
            resolved_ip_ttl_threshold,
            ping_reply_timeout
    );
}

/**
 * @brief Destructor.
 */
PingScheduler::~PingScheduler()
{
}

void PingScheduler::stop_pinging()
{
    Ping->stop_pinging();
}

/**
 * @brief Start into infinite loop of calls to resolve_and_ping
 *   IoService which is started here and can be stopped through stop_pinging
 */
void PingScheduler::start_pinging()
{
    EverHadAnyIP = false;

    if ( FirstDelay > 0 )
    {
        GlobalLogger.info() << "Delaying first ping by " << FirstDelay << "s";
        (void) NextPingTimer.expires_from_now( seconds( FirstDelay ) );
        NextPingTimer.async_wait( bind( &PingScheduler::resolve_and_ping, this,
                                              boost::asio::placeholders::error ) );
    }
    else
        resolve_and_ping(boost::system::error_code());
}


/**
 * @brief Check whether need to resolve host names via DNS and then
 *   call ping;
 *   If name resolution fails too often, reports this to HostAnalyzer
 */
void PingScheduler::resolve_and_ping(const boost::system::error_code &error)
{
    if ( error )
    {
        if ( error ==  boost::asio::error::operation_aborted )
            GlobalLogger.error() << "Timer for resolve_and_ping was cancelled! Stopping" << endl;
        else
            GlobalLogger.error() << "Received error " << error
                                 << " waiting for resolve_and_ping! Stopping" << endl;
        return;
    }

    bool ips_up_to_date = EverHadAnyIP && !Ping->expired_resolved_ip();

    // determine if address resolution is required
    if ( !ips_up_to_date )
    {
        ips_up_to_date = Ping->resolve_ping_address();    // try to resolve
        if ( ips_up_to_date )
        {
            EverHadAnyIP = true;
            HostAnalyzer.set_resolved_ip_count( Ping->get_resolved_ip_count() );
            AddressResolutionAttempts = 0;
        }
    }
    // (may still have no or only outdated IPs)


    if ( EverHadAnyIP )
    {   // have IPs --> can ping (even if IPs are out of date)
        if ( !ips_up_to_date )
            GlobalLogger.info() << "Using outdated IPs" << endl;
        ping();
    }
    else
    {   // no IPs --> try again later, possibly report offline before
        AddressResolutionAttempts++;
        if (AddressResolutionAttempts >= MaxAddressResolutionAttempts)
        {
            GlobalLogger.notice() << "No IPs after " << AddressResolutionAttempts << " DNS queries!";
            HostAnalyzer.report_dns_resolution_failure();
        }
        (void) NextAddressTimer.expires_from_now( seconds( AddressResolveIntervalInSec ) );
        NextAddressTimer.async_wait( bind( &PingScheduler::resolve_and_ping, this,
                                                 boost::asio::placeholders::error ) );
    }
}

/**
 * @brief call Ping::ping and schedule a call to ping_done_handler when finished
 */
void PingScheduler::ping()
{
    Ping->ping( boost::bind(&PingScheduler::ping_done_handler, this, _1) );
}


/**
 * @brief called when Ping::ping is done; calls functions to update
 *   statistics, ping interval and elapsed time;
 *   schedules a call to resolve_and_ping, thereby closing the loop
 */
void PingScheduler::ping_done_handler( const bool ping_success )
{
    // post-processing
    // You must call these 3 methods exactly in this order
    // TODO Fix this method, once it has a semantic dependency with the
    // update_ping_statistics method, because it depends on the PingAnalyzer
    // statistics to update the exceeded_ping_failed_limit
    HostAnalyzer.update_ping_statistics( ping_success );
    update_ping_interval();
    update_ping_elapsed_time();

    // schedule next ping
    (void) NextPingTimer.expires_from_now( seconds( PingIntervalInSec ) );
    NextPingTimer.async_wait( bind( &PingScheduler::resolve_and_ping, this,
                                         boost::asio::placeholders::error ) );
}

void PingScheduler::update_ping_interval()
{
    // have to ping more often?
    if ( HostAnalyzer.exceeded_ping_failed_limit() )
    {
        PingIntervalInSec.speed_up();

        GlobalLogger.debug() << "- Speeding up ping interval to: " << PingIntervalInSec << "s"
                << endl;
    }
    else
    {
        PingIntervalInSec.back_to_original();

        GlobalLogger.debug() << "- Stick to the original ping interval: " << PingIntervalInSec << "s"
                << endl;
    }
}

void PingScheduler::update_ping_elapsed_time()
{
    ptime now = microsec_clock::universal_time();
    time_resolution_traits_adapted64_impl::int_type elapsed_time_in_sec =
            (now - TimeSentLastPing).total_seconds();
    GlobalLogger.debug() << "- Time elapsed since last ping: " << elapsed_time_in_sec << "s" << endl;

    TimeSentLastPing = microsec_clock::universal_time();
}

